/*
  uip network configuration console
*/

#include "uip.h"

#include <stdio.h>
#include <string.h>

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "config.h"

#define SET_IP(ip, a,b,c,d) { ip[0] = (b<<8) | a; ip[1] = (d<<8) | c; }

/* various flags stored with the configuration */
#define CONFIG_FLAG_DHCP     0x01   /* use DHCP */

#define MAX_BUFFER  34
#define PROMPT      "> "

/* global variables */
static char buffer_data[MAX_BUFFER+1];  // plus space for terminating nul
static unsigned char buffer_idx = 0;

/* store config itself in eeprom */
config_t config EEMEM;

/* led blink diagnostics, currently used: */
/* 2 : register write test failed */
/* 3 : card init failed */
/* 4 : no card detected */
void config_console_fatal(const char *str, unsigned char n) {
  unsigned char j, i;

  printf_P(PSTR("Fatal error: "));
  printf_P(str);
  printf_P(PSTR("\nError code #%d, halting!\n"), n);

  while(1) {
    wdt_reset();      

    PORTB &= ~_BV(2);   // green led off
    for(i=0;i<100;i++) 
      _delay_ms(10);     // _delay_ms() is limited to 16ms @ 16Mhz

    for(j=0;j<n;j++) {
      PORTB |= _BV(2);   // green led on
      for(i=0;i<20;i++) 
	_delay_ms(10);

      PORTB &= ~_BV(2);   // green led off
      for(i=0;i<20;i++) 
	_delay_ms(10);
    }
  }
}

#ifdef CONSOLE_CONFIG_WLAN
/* routines to copy strings from and to eeprom */
void config_eeprom_write_str(const void *pointer_ram, void *pointer_eeprom) {
  eeprom_write_block(pointer_ram, pointer_eeprom, strlen(pointer_ram)+1);
}

void config_eeprom_read_str(char *pointer_ram, char *pointer_eeprom) {
  while(eeprom_read_byte(pointer_eeprom))
    *pointer_ram++ = eeprom_read_byte(pointer_eeprom++);

  *pointer_ram = 0;
}

unsigned char config_eeprom_strlen(char *pointer_eeprom) {
  unsigned char i = 0;
  while(eeprom_read_byte(pointer_eeprom++)) i++;
  return i;
}
#endif

/* --------------------- basic uart io functions ---------------- */
int _kbhit(void) {
  return UCSRA & _BV(RXC);
}

int readch(void) {
  loop_until_bit_is_set(UCSRA, RXC);
  return UDR;
}

void writech(int ch) {
  putchar(ch);
  fflush(stdout);
}

void writeln(char *line) {
  while(*line)
    writech(*line++);
}

/* -------------------------------------------------------------- */
static void help(void) {
  writeln("Supported commands:\n");
  writeln("?                  - show this help\n");
#ifdef CONSOLE_CONFIG_WLAN
  writeln("C num              - set channel (1..13)\n");
#endif
  writeln("D {on|off}         - enable/disable DHCP\n");
#ifdef CONSOLE_CONFIG_WLAN
  writeln("E essid            - enter essid\n");
#endif
  writeln("G addr             - set IP gateway\n");
  writeln("I addr             - set IP address\n");
#ifdef CONSOLE_CONFIG_WLAN
  writeln("K <key>            - set encryption key\n");
  writeln("M {managed|ad-hoc} - set mode\n");
#endif
  writeln("N addr             - set IP netmask\n");
  writeln("R                  - reboot\n");
  writeln("S                  - show config\n");
}

void error(char *msg) {
  writeln(msg);
  help();
}

/* skip white spaces */
static char *skip_white(char *in) {
  while((*in == ' ')&&(in))
    in++;

  return in;
}

static int get_ip(char *buffer, int len, uip_ipaddr_t v) {
  char *ip = (char*)v;
  int n = 0;

  /* skip leading spaces */
  while((*buffer == ' ')&&(len)) {
    buffer++; len--;
  }
    
  /* still data left? */
  if(len) {

    while(n != 4) {
      char cur = 0; 

      /* parse one decimal value */
      while((*buffer >= '0')&&(*buffer <= '9')&&(len)) {
	cur *= 10;
	cur += *buffer - '0';

	buffer++;
	len--;
      }

      /* write finished byte to eeprom */
      eeprom_write_byte(ip+n, cur);

      /* forth byte not reached */
      if(n<3) {
	if((!len)||(*buffer != '.')) {
	  error("Not a valid IP address\n");
	  return 0;
	}
	buffer++;
	len--;
      }

      n++;
    }

    return 1;
  }

  /* parsing failed if we got here ... */
  help();
  return 0;
}

void show_config(void) {
  char tmp[33];  // longest possible string

  printf_P(PSTR("Configuration:\n"));

#ifdef CONSOLE_CONFIG_WLAN
  printf_P(PSTR("WLAN mode is:    "));
  if(eeprom_read_byte(&config.flags) & CONFIG_FLAG_MANAGED)
    printf_P(PSTR("Managed\n"));
  else {
    printf_P(PSTR("Ad-Hoc\n"));
    printf_P(PSTR("Channel:         %d\n"), eeprom_read_byte(&config.channel));
  }

  config_eeprom_read_str(tmp, config.essid);
  printf_P(PSTR("ESSID:           %s\n"), tmp);

  printf_P(PSTR("Encryption is:   "));
  if(eeprom_read_byte(&config.flags) & (CONFIG_KEY_WEP40|CONFIG_KEY_WEP104)) {
    int i = (eeprom_read_byte(&config.flags) & CONFIG_KEY_WEP40)?40:104;
    char *p = config.key;

    printf_P(PSTR("%d bit WEP\n"), i);
    printf_P(PSTR("Encryption key:  "));

    i >>= 3;
    while(i--)
      printf_P(PSTR("%02x"), 0xff & eeprom_read_byte(p++));

    printf_P(PSTR("\n"));
  } else
    printf_P(PSTR("off\n"));
#endif

  printf_P(PSTR("DHCP is:         o"));
  if(eeprom_read_byte(&config.flags) & CONFIG_FLAG_DHCP) {
    printf_P(PSTR("n\n"));
  } else {
    uip_ipaddr_t ipaddr;
    printf_P(PSTR("ff\n"));

    eeprom_read_block(ipaddr, config.addr, 4);
    printf_P(PSTR("IP address:      %u.%u.%u.%u\n"), 
	 uip_ipaddr1(ipaddr), uip_ipaddr2(ipaddr),
	 uip_ipaddr3(ipaddr), uip_ipaddr4(ipaddr));
    eeprom_read_block(ipaddr, config.mask, 4);
    printf_P(PSTR("Network mask:    %u.%u.%u.%u\n"), 
	 uip_ipaddr1(ipaddr), uip_ipaddr2(ipaddr),
	 uip_ipaddr3(ipaddr), uip_ipaddr4(ipaddr));
    eeprom_read_block(ipaddr, config.gate, 4);
    printf_P(PSTR("Gateway address: %u.%u.%u.%u\n"), 
	 uip_ipaddr1(ipaddr), uip_ipaddr2(ipaddr),
	 uip_ipaddr3(ipaddr), uip_ipaddr4(ipaddr));
  }
}

#ifdef CONSOLE_CONFIG_WLAN
/* parse hex digit. required for encryption key parsing */
static int hexdigit(char c) {
  if((c >= '0') && (c <= '9')) return c - '0';
  if((c >= 'a') && (c <= 'f')) return c - 'a' + 10;
  if((c >= 'A') && (c <= 'F')) return c - 'A' + 10;
  return -1;
}
#endif

void config_console_handle(void) {
  if(_kbhit()) {
    char c = readch();

    switch(c) {
    case 8:
      if(buffer_idx > 0) {
	writeln("\010 \010");  // bs space bs
	buffer_idx--;
      } else
	writech('\007');  // bling!!
      break;

    case 10:
    case 13:
      writeln("\n");

      /* process command */
      if(buffer_idx > 0) {
	char *w, flags = eeprom_read_byte(&config.flags);

	buffer_data[buffer_idx] = 0;
	w = skip_white(buffer_data+1);
	buffer_idx -= w-buffer_data;

	/* parse first byte */
	switch(buffer_data[0]) {
	  unsigned long val;
	  int i, j;

#ifdef CONSOLE_CONFIG_WLAN
	case 'C':  /* set channel */
	case 'c':
	  i = atoi(w);
	  if((i >= 1)&&(i <= 13)) {
	    eeprom_write_byte(&config.channel, i);
	    show_config();
	  } else
	    help();
	  break;
#endif

	case 'D':  /* toggle dhcp */
	case 'd':
	  if(strcasecmp(w, "on") == 0) {
	    eeprom_write_byte(&config.flags, flags | CONFIG_FLAG_DHCP);
	    show_config();
	  } else if(strcasecmp(w, "off") == 0) {
	    eeprom_write_byte(&config.flags, flags & ~CONFIG_FLAG_DHCP);
	    show_config();
	  } else
	    help();
	  break;

#ifdef CONSOLE_CONFIG_WLAN
	case 'E':  /* enter essid */
	case 'e':
	  if(strlen(w) > CONFIG_MAX_SSID_SIZE)
	    w[CONFIG_MAX_SSID_SIZE] = 0;

	  config_eeprom_write_str(w, config.essid);
	  show_config();
	  break;
#endif //CONSOLE_CONFIG_WLAN
	  
	case 'G':  /* set IP gateway */
	case 'g':
	  if(get_ip(w, buffer_idx, config.gate))
	    show_config();
	  break;

	case 'I':  /* set IP address */
	case 'i':
	  if(get_ip(w, buffer_idx, config.addr))
	    show_config();
	  break;

#ifdef CONSOLE_CONFIG_WLAN
	case 'K':  /* enter key */
	case 'k':
	  /* check if entire key is hex */
	  for(j=i=0;i<strlen(w);i++)
	    if(! (((w[i] >= '0')&&(w[i] <= '9')) ||
		  ((w[i] >= 'a')&&(w[i] <= 'f')) ||
		  ((w[i] >= 'A')&&(w[i] <= 'F'))))
	      j = 1;

	  printf_P(PSTR("is not hex = %d\n"), j);
	  printf_P(PSTR("key len = %d\n"), i);

	  if(i == 0) {
	    flags &= ~(CONFIG_KEY_WEP40 | CONFIG_KEY_WEP104);
	    eeprom_write_byte(&config.flags, flags);
	  } else if((i == 5)||(i == 10) && (!j)) {
	    flags &= ~CONFIG_KEY_WEP104;
	    flags |=  CONFIG_KEY_WEP40;
	    eeprom_write_byte(&config.flags, flags);
	  } else if((i == 13)||(i == 26) && (!j)) {
	    flags &= ~CONFIG_KEY_WEP40;
	    flags |=  CONFIG_KEY_WEP104;
	    eeprom_write_byte(&config.flags, flags);
	  } else
	    help();

	  /* convert hex string if it's not not-hex and has correct length */
	  if((!j) && ((i == 10) || (i == 26))) {
	    char *d = config.key;

	    while(i) {
	      eeprom_write_byte(d++, (hexdigit(w[0])<<4) + hexdigit(w[1]));
	      w += 2; i -= 2;
	    }
	  }

	  /* directly copy ASCII string */
	  if((i == 5) || (i == 13)) 
	    eeprom_write_block(w, &config.key, i);

	  show_config();
	  break;
	  
	case 'M':  /* set mode */
	case 'm':
	  if(strcasecmp(w, "managed") == 0) {
	    flags |= CONFIG_FLAG_MANAGED;
	    eeprom_write_byte(&config.flags, flags);
	    show_config();
	  } else if((strcasecmp(w, "ad-hoc") == 0)||
		    (strcasecmp(w, "adhoc") == 0)) {
	    flags &= ~CONFIG_FLAG_MANAGED;
	    eeprom_write_byte(&config.flags, flags);
	    show_config();
	  } else
	    help();
	  break;
#endif //CONSOLE_CONFIG_WLAN

	case 'N':  /* set IP netmask */
	case 'n':
	  if(get_ip(w, buffer_idx, config.mask))
	    show_config();
	  break;

	case 'R':  /* reboot */
	case 'r':
	  printf_P(PSTR("Rebooting ...\n"));
	  wdt_enable(WDTO_15MS);  
	  for(;;);
	  break;

	case 'S':  /* show config */
	case 's':
	  show_config();
	  break;

	case '?':
	  help();
	  break;

	default:
	  printf_P(PSTR("Error, unknown command %c\n"), buffer_data[0]);
	  help();
	  break;
	}
      } else
	help();

      writeln(PROMPT);
      buffer_idx = 0;
      break;

    default:
      if((c >=32) && (c < 127)) {
	if(buffer_idx < MAX_BUFFER-1) {
	  buffer_data[buffer_idx++] = c;
	  writech(c);
	}
      }
      break;
    }
  }
}

#define DEFAULT_FLAGS (CONFIG_FLAG_DHCP | CONFIG_FLAG_MANAGED)

void config_init(void) {
  /* set defaults */
  if(eeprom_read_byte(&config.valid) != 0x42) {
    uip_ipaddr_t tmp;

    printf_P(PSTR("No default config, creating one ...\n"));

    eeprom_write_byte(&config.valid, 0x42);

#ifdef CONSOLE_CONFIG_WLAN
    /* some default WLAN setting */
    config_eeprom_write_str("Tantil", config.essid);
    eeprom_write_byte(&config.flags, CONFIG_KEY_WEP40 | DEFAULT_FLAGS);
    /* the following string is the key c0ffee1234 in octal notation */
    config_eeprom_write_str("\300\377\356\022\064", config.key);
    eeprom_write_byte(&config.channel, 11);
#else
    eeprom_write_byte(&config.flags, DEFAULT_FLAGS);
#endif

    SET_IP(tmp, 192,168,1,123);
    eeprom_write_block(tmp, config.addr, 4);
    SET_IP(tmp, 255,255,255,0);
    eeprom_write_block(tmp, config.mask, 4);
    SET_IP(tmp, 192,168,1,254);
    eeprom_write_block(tmp, config.gate, 4);
  }
}


